/*
This is the core library to control a Yaesu FT817 transceiver. 
It uses the Arduino Serial.* methods to access the device (mapped via "FT817_Serial").

Copyright (c) 2017 OE8WOZ.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution (including sources).

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

#include "YaesuControl.h"

YaesuControl::YaesuControl()
{
  baudrate = FT817_BAUD48;
}

// =================================================
// routines handling the FT-817 communication
// =================================================

// set up serial port
// (with or without default baudrate)
int16_t YaesuControl::begin(uint16_t baud)
{ 
  baudrate = baud;
  return begin();
}
int16_t YaesuControl::begin()
{
  FT817_Serial.begin(baudrate, SERIAL_8N2);

  return 0;
}

// start communication with a FT-817
// we check that RX is high prior to communication
// this may help to determine if a FT-871 is actually
// connected before trying to do a readout
// returns -1 if no FT-817 was found
int16_t YaesuControl::init()
{
  // we don't check on Serial0 (and thus UNO boards) as there is
  // always the USB interface on the board connected (thus RX high)
  #if FT817_Serial==Serial1
    pinMode(19, INPUT); // just in case, should not have any effect
    if (digitalRead(19)==LOW) return -1;
  #else
    #if FT817_Serial==Serial2
      pinMode(17, INPUT); // just in case, should not have any effect
      if (digitalRead(17)==LOW) return -1;
    #else
      #if FT817_Serial==Serial3
        pinMode(15, INPUT); // just in case, should not have any effect
        if (digitalRead(15)==LOW) return -1;
      #else
        #warning "Please note: the FT817 uses the Arduino debug Serial!"
      #endif
    #endif
  #endif

  // we wait a bit - in case someone use the init in a loop to
  // ensure the FT-817 gets time to start up
  delay(1000);

  // try to read the QRG and mode
  uint32_t f;
  if (getFreqMode(&f)<0) return -1;

  // seems to be ok, continue
  return 0;
}

// send single-byte command to FT-817
// (command without parameters)
void YaesuControl::sendCommand(uint8_t command) {
  uint8_t cmd[] = {FT817_EMPTY_PARAM,FT817_EMPTY_PARAM,FT817_EMPTY_PARAM,FT817_EMPTY_PARAM,command};
  // packet must be sent with max. 200ms between bytes
  FT817_Serial.write(cmd,5);
  Serial.println();
  Serial.print(command, HEX);
}

// send command with several bytes to FT-817
// (format is 4 byte parameters, 1 byte command)
void YaesuControl::sendLongCommand(uint8_t command[]) {
  // packet must be sent with max. 200ms between bytes
  FT817_Serial.write(command,5);
  Serial.println();
  Serial.print(command[4], HEX);
}

// special serial read of FT-817 with timeout of 200ms
// (prevent stall when e.g. no FT-817 is connected)
int16_t YaesuControl::readByte() {
  uint32_t now = millis();
  while (!FT817_Serial.available()) {
    if ((millis()-now) >= 200) return -1;
  }
  Serial.print(" ");
  Serial.print(FT817_Serial.available());
  return FT817_Serial.read();
}

// read actual EEP content of the FT-817
int32_t YaesuControl::getEEP(uint16_t adr) {
  uint8_t cmd[] = {(uint8_t)(adr>>8),(uint8_t)(adr&0xFF),FT817_EMPTY_PARAM,FT817_EMPTY_PARAM,FT817_CMD_EEP_READ};  
  sendLongCommand(cmd);
  return readByte() | (readByte()<<8);
}

// set EEP content of the FT-817, we are really carefully here !
uint8_t YaesuControl::setEEP(uint16_t adr, uint8_t mask, uint8_t val) {
  // prevent modification of factory settings
  if (adr<0x54) return -1;

  // read content twice
  uint8_t rdcmd[] = {(uint8_t)(adr>>8),(uint8_t)(adr&0xFF),FT817_EMPTY_PARAM,FT817_EMPTY_PARAM,FT817_CMD_EEP_READ};  
  sendLongCommand(rdcmd);
  int16_t byte1 = readByte();
  int16_t byte2 = readByte();
  sendLongCommand(rdcmd);
  int16_t byte3 = readByte();
  int16_t byte4 = readByte();

  // fail on read - we are done!
  if ((byte1==-1) || (byte2==-1) || (byte1!=byte3) || (byte2!=byte4)) {
    return -1;
  }

  // update content
  byte1 = (byte1 & mask ) | val;
  uint8_t wrcmd[] = {(uint8_t)(adr>>8),(uint8_t)(adr&0xFF),(uint8_t)byte1,(uint8_t)byte2,FT817_CMD_EEP_WRITE};  
  sendLongCommand(wrcmd);

  // return the byte we will receive here (usually 0)
  return readByte();
}

// check the RX status of the FT-817
int16_t YaesuControl::getRxStatus() {
  sendCommand(FT817_CMD_RX_READ);
  return readByte();
}

// check the TX status of the FT-817
int16_t YaesuControl::getTxStatus() {
  sendCommand(FT817_CMD_TX_READ);
  return readByte();
}

// get the FT-817 configuration
int16_t YaesuControl::getConfig(uint8_t *buf) {
  int i;
  sendCommand(FT817_CMD_CONF_READ);
  for(i=0;i<9;i++) {
    int16_t c = readByte();
    if (c<0) return -1;
    *buf++ = (uint8_t) c;
  }
  return i;
}

// send command to read frequency/mode from FT-817
int16_t YaesuControl::getFreqMode(uint32_t *freq) {
  int16_t b;

  sendCommand(FT817_CMD_FREQ_READ);

  // this is the frequency
  uint8_t byte[4];
  for (int i = 0; i < 4; i++) {
    b = readByte();
    if (b<0) return -1;
    byte[i] = (uint8_t)(b&0xFF); 
  }
  *freq = bcd2bin(byte);

  // this is the mode
  b = readByte();
  if (b<0) return -1;
  uint8_t mode = b&0xFF;

  return mode;
}

// send command to set mode on FT-817
// (modes defined as FT817_MODE_*, WFM can not be set this way)
void YaesuControl::setMode(uint8_t mode) {
  uint8_t cmd[] = {mode,FT817_EMPTY_PARAM,FT817_EMPTY_PARAM,FT817_EMPTY_PARAM,FT817_CMD_MODE_WRITE};  
  sendLongCommand(cmd);
  readByte(); // returns zero
}

// send command to set frequency on FT-817
// (will do band switching automatically)
void YaesuControl::setFreq(uint32_t freq) {
  uint8_t byte[4];
  bin2bcd(freq,byte);
  uint8_t cmd[] = {byte[3],byte[2],byte[1],byte[0],FT817_CMD_FREQ_WRITE};
  sendLongCommand(cmd);
  readByte(); // returns zero
}


// send command to set front/rear port on FT-817
// (1...rear port)
void YaesuControl::setPort(uint8_t port) {
  setEEP(0x7A,~0x30,port?0x30:0x00);
}

// send command to get front/rear port from FT-817
// (1...rear port)
uint8_t YaesuControl::getPort(void) {
  return (getEEP(0x7A)&0x30)?1:0;
}

// send command to set power on FT-817
// (3...0.5/2...1.0/1...2.5 W/0...5W)
void YaesuControl::setPwr(uint8_t pwr) {
  setEEP(0x79,~0x03,pwr&0x03);
  readByte(); // why do we get here another byte...?
}

// send command to get power from FT-817
// (3...0.5/2...1.0/1...2.5 W/0...5W)
uint8_t YaesuControl::getPwr(void) {
  return (getEEP(0x79)&0x03);
}

// send command to set PTT on FT-817
// (1...on, 0...off)
uint8_t YaesuControl::setPTT(uint8_t ptt) {
  if (ptt)
    sendCommand(FT817_CMD_PTT_ON);
  else
    sendCommand(FT817_CMD_PTT_OFF);
  return (readByte()==0xF0?1:0);
}

// send command to get PTT from FT-817
// (1...on, 0...off)
uint8_t YaesuControl::getPTT(void) {
  sendCommand(FT817_CMD_PTT_READ);
  return (readByte()==0xF0?1:0);
}

// decode an number in Hz we get from the FT-817 in a special format
// (4 bytes == 8 digits, XXX.XXX.XX0 Hz)
uint32_t YaesuControl::bcd2bin(unsigned char const* bcd)
{
  uint32_t val = 0;
  uint8_t len=4;
  while (len--) {
      val += (*bcd>>4);
      val *= 10;
      val += (*bcd&15);
      val *= 10;
      ++bcd;
  }
  return  val;
}

// encode an number wesend to the FT-817 in a special format
// (4 bytes == 8 digits, XXX.XXX.XX0 Hz)
void YaesuControl::bin2bcd(uint32_t val, unsigned char *bcd)
{ uint8_t len = 4;
  while (len--) {
    val=val/10L;
    *bcd = val%10L;
    val=val/10L;
    *bcd |= ((val%10L)<<4);
    ++bcd;
  }
}
