/*
This is the core library to control a Yaesu FT817 transceiver. 
It uses the Arduino Serial.* methods to access the device (mapped via "FT817_Serial").

Copyright (c) 2017 OE8WOZ.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution (including sources).

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef YAESUCONTROL_H
#define YAESUCONTROL_H

#if ARDUINO >= 100
  #include "Arduino.h"
#endif

#define FT817_LIB_VERSION 101

// FT-817 UART speeds

#define FT817_BAUD48 4800
#define FT817_BAUD96 9600
#define FT817_BAUD384 38400

// try specify serial to use if we don't use a simple Uno board
#ifndef AVR_UNO
#define FT817_Serial Serial1
#endif

// FT-817 commands (official in FT-817 manual) -> not used yet

#define FT817_CMD_LOCK_ON  0x00    //(no parameter)
#define FT817_CMD_LOCK_OFF 0x80    //(no parameter)

#define FT817_CMD_CLAR_ON  0x05    //(no parameter)
#define FT817_CMD_CLAR_OFF 0x85    //(no parameter)

#define FT817_CMD_CLAR_WRITE 0xF5  //(P1: +/- offset, P3-P4: BCD frequency)

#define FT817_CMD_VFO_TOGGLE 0x81  //(no parameter)

#define FT817_CMD_SPLIT_ON  0x02   //(no parameter)
#define FT817_CMD_SPLIT_OFF 0x82   //(no parameter)

#define FT817_CMD_SHIFT_WRITE 0x09 //(P1: -/+/off)
#define FT817_CMD_OFF_WRITE   0xF9 //(P1-P4: BCD frequency)

#define FT817_CMD_CSSDCS_WRITE 0x0A //(P1: DCS/CTCSS/ENC/off)
#define FT817_CMD_CTCSS_WRITE  0x0B //(P1-P2: BCD tone)
#define FT817_CMD_DCS_WRITE    0x0C //(P1-P2: BCD code)

#define FT817_CMD_PWR_ON  0x0F      //(no parameter)
#define FT817_CMD_PWR_OFF 0x8F      //(no parameter)

// FT-817 commands (official in FT-817 manual)

#define FT817_CMD_PTT_ON   0x08    //(no parameter)    --> 1 byte
#define FT817_CMD_PTT_OFF  0x88    //(no parameter)    --> 1 byte

#define FT817_CMD_FREQ_WRITE 0x01  //(P1-P4: BCD frequency)    --> 1 byte
#define FT817_CMD_MODE_WRITE 0x07  //(P1:    modulation mode)  --> 1 byte

#define FT817_CMD_RX_READ   0xE7    //(no parameter)  --> 1 byte
#define FT817_CMD_TX_READ   0xF7    //(no parameter)  --> 1 byte

#define FT817_CMD_FREQ_READ 0x03    //(no parameter)  --> 5 byte

// FT-817 commands (inofficial from web: http://www.ka7oei.com/ft817_meow.html)

#define FT817_CMD_PTT_READ  0x10    //(no parameter)

#define FT817_CMD_EEP_READ  0xBB
#define FT817_CMD_EEP_WRITE 0xBC

// FT-817 radio config

#define FT817_CMD_CONF_READ  0xA7    //(no parameter) --> 9 byte

// FT-817 empty parameter

#define FT817_EMPTY_PARAM 0x00

// FT-817 modes

#define FT817_MODE_LSB 0x00
#define FT817_MODE_USB 0x01
#define FT817_MODE_CW  0x02
#define FT817_MODE_CWR 0x03
#define FT817_MODE_AM  0x04
#define FT817_MODE_WFM 0x06
#define FT817_MODE_FM  0x08
#define FT817_MODE_DIG 0x0A
#define FT817_MODE_PKT 0x0C

// FT-817 class declaration and methods

class YaesuControl
{
  public:
    YaesuControl(void);

    int16_t begin(uint16_t baudrate);
    int16_t begin();

    int16_t init();

    int32_t getEEP(uint16_t adr);

    int16_t getRxStatus();
    int16_t getTxStatus();

    int16_t getConfig(uint8_t *buf);

    void setFreq(uint32_t freq);
    void setMode(uint8_t mode);
    int16_t getFreqMode(uint32_t *freq);

    void setPort(uint8_t port);
    uint8_t getPort(void);

    void setPwr(uint8_t pwr);
    uint8_t getPwr(void);

    uint8_t setPTT(uint8_t ptt);
    uint8_t getPTT(void);

  private:

    uint32_t baudrate;

    void sendCommand(uint8_t command);
    void sendLongCommand(uint8_t command[]);
    int16_t readByte(void);

    uint8_t setEEP(uint16_t adr, uint8_t mask, uint8_t val);

    uint32_t bcd2bin(uint8_t const* bcd);
    void bin2bcd(uint32_t val, uint8_t *bcd);
};

#endif //YAESUCONTROL_H